/* ------------------------------------------------------------------------- *\
 * NOM
 *      bst
 * SYNOPSIS
 *      bst [-w] database.csv
 * DESCIRPTION
 *      Loads the data into the BST
 * OPTION
 *      -w      Whether to load the whole file as one. If not present
 *              the element are inserted one by one.
 * USAGE
 *      ./best database.csv
 \* ------------------------------------------------------------------------- */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdbool.h>
#include <time.h>
#include "BinarySearchTree.h"

// Max length of a single line in a .csv file
static const size_t LINE_SIZE = 512;
// Whether to build the whole BST at once
static const char* WHOLE_OPT = "-w";

int main(int argc, char** argv)
{
    // Parses command line input
    char* fileName = NULL;
    unsigned int index = 1;
    bool whole = false;
    if (argc < 2)
    {
        fprintf(stderr, "%s is expecting at least one parameter (.csv file)\n", argv[0]);
        exit(EXIT_FAILURE);
    }
    if (argc == 3)
    {
        if (strcmp(argv[index], WHOLE_OPT) == 0)
        {
            whole = true;
            index++;
        }
        else
        {
            fprintf(stderr, "Unrecognized option '%s'\n", argv[index]);
            exit(EXIT_FAILURE);
        }
    }
    fileName = argv[index];

    // Opens the file
    FILE* fileObj = fopen(fileName, "r");
    if (fileObj == NULL)
    {
        fprintf(stderr, "Could not open file '%s'. Exiting...\n", fileName);
        exit(EXIT_FAILURE);
    }

    // Counts the amount of lines
    unsigned int nbLines = 0;
    while(!feof(fileObj))
    {
        char ch = fgetc(fileObj);
        if(ch == '\n')
            nbLines++;
    }
    rewind(fileObj);

    // Allocates "keys" and "values" arrays
    double* keys = malloc(sizeof(double) * nbLines);
    if (!keys)
    {
        fprintf(stderr, "Allocation error: key. Exiting...\n");
        exit(EXIT_FAILURE);
    }
    char** values = malloc(sizeof(char*) * nbLines);
    if (!values)
    {
        fprintf(stderr, "Allocation error: values. Exiting...\n");
        exit(EXIT_FAILURE);
    }
    for(unsigned int i = 0; i < nbLines; i++)
    {
        keys[i] = 0.0;
        values[i] = NULL;
    }

    // Start reading the file, line by line (N.B.: fgets() reads until next \n or EOF character)
    char line[LINE_SIZE];
    size_t nbLine = 0;
    char delim = ';';
    char end = '\n';
    while (fgets(line, sizeof(line), fileObj) != NULL)
    {
        // Finds position of ";" and "\n"
        unsigned short delimPos = 0, endPos = 0;
        for(unsigned short i = 0; i < LINE_SIZE && endPos == 0; i++)
        {
            if(line[i] == delim)
                delimPos = i;
            if(line[i] == end)
                endPos = i;
        }

        // Product name
        values[nbLine] = malloc(sizeof(char) * (delimPos + 1));
        if (!values[nbLine])
        {
            fprintf(stderr, "Allocation error: value at line %zu. Exiting...\n", nbLine);
            exit(EXIT_FAILURE);
        }
        memcpy(values[nbLine], line, delimPos);
        values[nbLine][delimPos] = '\0';

        // Price (key)
        char* keyStr = malloc(sizeof(char) * endPos - delimPos);
        if (!keyStr)
        {
            fprintf(stderr, "Allocation error: price at line %zu. Exiting...\n", nbLine);
            exit(EXIT_FAILURE);
        }
        memcpy(keyStr, line + delimPos + 1, endPos - delimPos - 1);
        keyStr[endPos - delimPos - 1] = '\0';
        keys[nbLine] = atof(keyStr);
        free(keyStr);

        nbLine++;
    }
    fclose(fileObj);

    clock_t startBuilding = clock();
    BinarySearchTree* bst;
    if (whole)
    {
        bst = createBinarySearchTree((const double*)keys, (const char**)values, nbLines);
        if (!bst)
        {
            fprintf(stderr, "Allocation error: Whole BST creation. Exiting...\n");
            exit(EXIT_FAILURE);
        }
    }
    else
    {
        bst = createEmptyBinarySearchTree();
        if (!bst)
        {
            fprintf(stderr, "Allocation error: Empty BST creation. Exiting...\n");
            exit(EXIT_FAILURE);
        }
        for (unsigned int i = 0; i < nbLines; i++)
        {
            int res = insertElement(bst, keys[i], values[i]);
            if(!res)
            {
                fprintf(stderr, "Allocation error: new BST element. Exiting...\n");
                exit(EXIT_FAILURE);
            }
        }
    }
    double buildingTime = ((double) (clock() - startBuilding)) / CLOCKS_PER_SEC;
    printf("Tree of %zu elements (height of %d) built in %.3f seconds\n", getNumElements(bst), getHeight(bst), buildingTime);

    // Some test code
    printf("For 349.90 euros, you can get %s\n", findElement(bst, 349.90));
    printf("For 19.90 euros, you can get %s\n", findElement(bst, 19.90));
    printf("For 339.90 euros, you can get %s\n", findElement(bst, 339.90));
    printf("For 44.10 euros, you can get %s\n", findElement(bst, 44.10));
    printf("For 74.82 euros, you can get %s\n", findElement(bst, 74.82));

    // Another test code
    printf("Removing %s...\n", removeElement(bst, 29.90));
    printf("Removing %s...\n", removeElement(bst, 269.99));
    printf("Removing %s...\n", removeElement(bst, 37.90));
    printf("Removing %s...\n", removeElement(bst, 389.99));
    printf("Removing %s...\n", removeElement(bst, 208.74));
    printf("The tree now contains %zu elements.\n", getNumElements(bst));

    // Two tests for getElementsInRange()
    char** valuesRange;
    size_t nbValuesRange = getElementsInRange(bst,10.00, 40.00, (const char***) &valuesRange);
    printf("\nIn range [10, 40]:\n");
    for(unsigned int i = 0; i < nbValuesRange; i++)
        printf("%s\n", valuesRange[i]);
    free(valuesRange);

    char** valuesRange2;
    size_t nbValuesRange2 = getElementsInRange(bst, 300.00, 400.00, (const char***) &valuesRange2);
    printf("\nIn range [300, 400]:\n");
    for(unsigned int i = 0; i < nbValuesRange2; i++)
        printf("%s\n", valuesRange2[i]);
    free(valuesRange2);

    freeBinarySearchTree(bst);
    free(keys);
    for(unsigned int i = 0; i < nbLines; i++)
        free(values[i]);
    free(values);

    exit(EXIT_SUCCESS);
}
